/*
 
  copyright (c) 2004, Michal Salaban <emes@pld-linux.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2 as
  published by the Free Software Foundation (see file COPYING for details).

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: parse.c,v 1.6 2004/02/08 18:05:54 emes Exp $
 
*/

#include "lex.h"
#include "../exit.h"
#include "../xalloc.h"
#include "../conditions.h"
#include "parse.h"
#include <stdio.h>
#include <stdlib.h>

#define ERRMSGBUF 128


extern  stat_la1,
        stat_la5,
        stat_la15,
        stat_memfree,
        stat_memused,
        stat_swapfree,
        stat_swapused;


#ifdef DEBUG
void * parse_error (tokdata *tok, const char *func, const char *desc)
#else
void * parse_error (tokdata *tok)
#endif
{
    char    *buf = (char *) xalloc (NULL, ERRMSGBUF);
    
#ifdef DEBUG
    snprintf (
            buf,
            ERRMSGBUF - 1,
            "at line %d before token %s, reported by function %s [%s]",
            tok->line,
            tok2string (tok->type),
            func,
            desc
            );
#else
    snprintf (
            buf,
            ERRMSGBUF - 1,
            "at line %d",
            tok->line
            );
#endif
    
    bail_out (EXIT_PARSE, buf);
}

parserdata * get_logdata (tokdata *tok)
{
    /*  grammar:
     *  
     *  logdata := log [once]
     *
     */
    parserdata  *res;
    logdata     *log;   /* helper */
    
    if (tok->type != KW_LOG) return NULL;
    
    tok ++;

    /* that's log data. initialize structures */
    res = (parserdata *) xalloc (NULL, sizeof (parserdata));
    res->parsed = xalloc (NULL, sizeof ( logdata));
    log = (logdata *) res->parsed;
    if (tok->type == KW_ONCE) {
        log->once = 1;
        tok ++;
    } else log->once = 0;

    res->ptr = tok;
    return res;
}

parserdata * get_rundata (tokdata *tok)
{
    /*  grammar:
     *
     *  rundata := run [once] "command"
     *
     */
#ifdef DEBUG
    const char  *thisfuncname = "get_rundata()";
#endif
    parserdata  *res;
    rundata     *run;   /* helper */

    if (tok->type != KW_RUN) return NULL;

    tok ++;

    /* that's run data. initialize structures */
    res = (parserdata *) xalloc (NULL, sizeof (parserdata));
    res->parsed = xalloc (NULL, sizeof (rundata));
    run = (rundata *) res->parsed;
    if (tok->type == KW_ONCE) {
        run->once = 1;
        tok ++;
    } else run->once = 0;

    if (tok->type != VA_STR) 
#ifdef DEBUG
        parse_error (tok, thisfuncname, "run arg not string");
#else
        parse_error (tok);
#endif
    run->runcmd = tok->val;
    tok ++;

    res->ptr = tok;
    return res;
}

parserdata * expression_decompose (parserdata *left)
{
#ifdef DEBUG
    const char  *thisfuncname = "expression_decompose()";
#endif
    expression      *resexp;
    parserdata      *right, *res;
    tokdata         *tok = left->ptr;
    
    if ((tok->type == LO_OR) || (tok->type == LO_AND)) {
        /* this is compound expression */
        res = (parserdata *) xalloc (NULL, sizeof (parserdata));
        resexp = (expression *) xalloc (NULL, sizeof (expression));
        res->parsed = (void *) resexp;
        resexp->type = RELATION;
        /* set the expression we found before as first argument */
        resexp->arg1 = left->parsed;
        /* set operator */
        resexp->op = tok->type;
        tok ++;
        /* get next expression as second argument.
         * there must be one (we had an operator).
         */
        right = get_expression (tok);
        if (right == NULL)
#ifdef DEBUG
        parse_error (tok, thisfuncname, "no right arg");
#else
        parse_error (tok);
#endif
        /* second argument is ok */
        resexp->arg2 = right->parsed;
        res->ptr = right->ptr;

        return res;
        
    } else return left; /* we had no operator */
}

parserdata * get_expression (tokdata *tok)
{
    /*  grammar:
     *
     *  condition := <atomic> [<op> <expression>]
     *             | <block_expression> [<op> <expression>]
     */
#ifdef DEBUG
    const char  *thisfuncname = "get_expression()";
#endif
    
    parserdata      *left;

    left = get_block_expression (tok);
    if (left != NULL) /* this is a block */
        return expression_decompose (left);
    
    left = get_atomic_expression (tok);
    if (left != NULL) /* this is an atomic rule */
        return expression_decompose (left);        
    
#ifdef DEBUG
    parse_error (tok, thisfuncname, "neither block nor atomic");
#else
    parse_error (tok);
#endif
}

parserdata * get_atomic (tokdata *tok)
{
#ifdef DEBUG
    const char  *thisfuncname = "get_atomic()";
#endif
    atomic      *at = (atomic *) xalloc (NULL, sizeof (atomic));
    tokdata     *var = tok,
                *op = tok + 1,
                *val = tok + 2;
    parserdata  *res;
    double      *tmp;
                
    tok += 3;

    switch (var->type) {
        case ID_LA1:
            at->type = DOUBLE;
            at->op = op->type;
            at->val = &stat_la1;
            break;
        case ID_LA5:
            at->type = DOUBLE;
            at->op = op->type;
            at->val = &stat_la5;
            break;
        case ID_LA15:
            at->type = DOUBLE;
            at->op = op->type;
            at->val = &stat_la15;
            break;
        case ID_MFREE:
            at->type = SIZE;
            at->op = op->type;
            at->val = &stat_memfree;
            break;
        case ID_MUSED:
            at->type = SIZE;
            at->op = op->type;
            at->val = &stat_memused;
            break;
        case ID_SFREE:
            at->type = SIZE;
            at->op = op->type;
            at->val = &stat_swapfree;
            break;
        case ID_SUSED:
            at->type = SIZE;
            at->op = op->type;
            at->val = &stat_swapused;
            break;
        default:
            return NULL;
    }
    
    switch (op->type) {
        /* allow only operators */
        case CO_EQ:
        case CO_NEQ:
        case CO_LE:
        case CO_LT:
        case CO_GE:
        case CO_GT:
            break;
        default:
#ifdef DEBUG
            parse_error (tok, thisfuncname, "invalid op");
#else
            parse_error (tok);
#endif
    }

    /* check value type */
    if (val->type == VA_INT) {
        /* this is integer. convert it to type required by label */
        if (at->type == SIZE) val->type = VA_SIZ;
        else if (at->type == DOUBLE) {
            /* long int 2 double */
            val->type = VA_DBL;
            tmp = xalloc (NULL, sizeof (double));
            *tmp = (double) *((long int *) val->val);
            free (val->val);
            val->val = tmp;
        }
    }
        
    if (((at->type == SIZE) && (val->type != VA_SIZ)) ||
         ((at->type == DOUBLE) && (val->type != VA_DBL)))
#ifdef DEBUG
        parse_error (val, thisfuncname, "type mismatch");
#else
        parse_error (val);
#endif

    /* copy threshold value */
    at->thresh = val->val;

    res = (parserdata *) xalloc (NULL, sizeof (parserdata));
    res->ptr = tok;
    res->parsed = (void *) at;
    return res;
}

parserdata * get_atomic_expression (tokdata *tok)
{
    /* get atomic and encapsulate into expression with type = ATOMIC */
    parserdata  *at = get_atomic (tok);
    expression  *ex;

    if (at == NULL) return NULL;
    ex = (expression *) xalloc (NULL, sizeof (expression));
    /* we'll use the same parserdata structure */

    ex->type = ATOMIC;
    ex->arg1 = at->parsed;
    at->parsed = ex;
    return at;
}

parserdata * get_block_expression (tokdata *tok)
{
    /*  grammar:
     * 
     *  block_expression := '{' <expression> '}'
     *
     */

#ifdef DEBUG
    const char  *thisfuncname = "get_block_expression()";
#endif
    
    parserdata      *expr;

    /* simply pass the expression enclosed in block.
     * check if block is well formated.
     */
    if (tok->type != BL_BEGIN) return NULL;
    tok ++;
    expr = get_expression (tok);
    if (expr == NULL) return NULL;
    if (expr->ptr->type != BL_END)
#ifdef DEBUG
        parse_error (tok, thisfuncname, "block not closed");
#else
        parse_error (tok);
#endif

    expr->ptr ++;
    return expr;
}

parserdata * get_ruleset (tokdata *tok)
{
    /*  grammar:
     *
     *  ruleset := 'if' <block_condition> <rundata> [<logdata>]
     *          |  'if' <block_condition> <logdata> [<rundata>]
     */

#ifdef DEBUG
    const char  *thisfuncname = "get_ruleset()";
#endif

    /* our result data */
    parserdata  *res = (parserdata *) xalloc (NULL, sizeof (parserdata)),
                *tmp, *run, *log;
    ruleset     *rlst;  /* helper */

    /* the only way to return w/o error is to find end-of-file */
    if (tok->type == END) return NULL;
    
    /* create ruleset structure */
    res->parsed = (ruleset *) xalloc (NULL, sizeof (ruleset));
    rlst = (ruleset *) res->parsed;
    /* reset fields */
    rlst->expr = NULL;
    rlst->run = NULL;
    rlst->log = NULL;
    rlst->hit = 0;
    rlst->prevhit = 0;
    
    /* ruleset begins with IF */
    if (tok->type != KW_IF)
#ifdef DEBUG
        parse_error (tok, thisfuncname, "no if keyword");
#else
        parse_error (tok);
#endif
    else tok ++;

    /* next should be expression in block */
    tmp = get_block_expression (tok);
    if (tmp == NULL)
#ifdef DEBUG
        parse_error (tok, thisfuncname, "no expr block");
#else
        parse_error (tok);
#endif

    /* save expression and skip to next token */
    rlst->expr = (expression *) tmp->parsed;
    tok = tmp->ptr;

    run = get_rundata (tok);
    log = get_logdata (tok);
    /* no log or run data? */
    if ((run == NULL) && (log == NULL))
#ifdef DEBUG
        parse_error (tok, thisfuncname, "no action");
#else
        parse_error (tok);
#endif

    /* this is full ruleset now but we try to get next action data */
    if (run != NULL) {
        /* save rundata && get logdata */
        rlst->run = (rundata *) run->parsed;
        log = get_logdata (run->ptr);

        /* if we have log sequence, save it.
         * skip to next token
         */
        if (log != NULL) {
            tok = log->ptr;
            rlst->log = (logdata *) log->parsed;
        } else tok = run->ptr;
    } else {
        rlst->log = (logdata *) log->parsed;
        /* save logdata && get rundata */
        run = get_rundata (log->ptr);

        if (run != NULL) {
            tok = run->ptr;
            rlst->run = (rundata *) run->parsed;
        } else tok = log->ptr;
    }

    /* ok, now we got full ruleset structure.
     * we need only to skip to next token.
     */
    
    res->ptr = tok;
    return res;
}

/* $Id: parse.c,v 1.6 2004/02/08 18:05:54 emes Exp $ */
